import numpy as np
import sys
import os
import re
import h5py
import quadpy as qd
from rho_calcu import *
import multiprocessing
import tqdm
import time

from pycuda import driver, compiler, autoinit
import pycuda.gpuarray as gpuarray
import skcuda.linalg as culi
import skcuda.misc as cumi
from pycuda.elementwise import ElementwiseKernel
from skcuda import cublas
from scipy.spatial.transform import Rotation as rot3d

cusqrt = ElementwiseKernel(
    "float *x, float *r",
    "r[i] = sqrt(x[i])",
    "cuda_sqrt")

def pycuda_sqrt(a_gpu):
  b_gpu = gpuarray.empty_like(a_gpu)
  cusqrt(a_gpu, b_gpu)
  return b_gpu

def trs_lgd(g_lgd_o):
  return np.divide(g_lgd_o, 1-g_lgd_o)

def gen_grid():
  # r
  lgd = qd.line_segment.gauss_legendre(100)
  g_lgd, w_lgd = lgd.points, lgd.weights
  # change from (-1,1) to (0,1) for weight, to (0,+inf) for grid
  g_lgd_o = (g_lgd+1)/2
  g_lgd = trs_lgd(g_lgd_o)
  w_lgd *= 0.5

  # theta & phi
  lbd = qd.sphere.lebedev_019()
  g_lbd, w_lbd = lbd.points, lbd.weights
  
  grid = np.outer(g_lgd, g_lbd).reshape(-1,3)
  wgt = np.outer(w_lgd, w_lbd).reshape(-1)

  return g_lgd_o, g_lgd, grid, wgt

def reduce_coord(posc, exp, num_prt):
  # construct tile matrx
  epsi = np.tile(exp, (num_prt,1))
  epsj = epsi.T
  # weighted exponents
  eps = np.add(epsi, epsj)
  ks = np.divide(epsi * epsj, eps)
  
  epsi_ci = np.tile(np.multiply(posc, exp[:,np.newaxis]), (num_prt,1,1))
  epsj_cj = np.transpose(epsi_ci, (1,0,2))
  # weighted position of center
  wposc = np.divide(np.add(epsi_ci, epsj_cj), eps[:,:,np.newaxis])

  return eps, ks, wposc

def calcu_prtxprt_exp(g_lgd, rpt, eps_ij):
  prtxprt_exp = np.exp(-eps_ij*np.power(g_lgd,2))
  # repeat rpt times for all spherical terms 
  prtxprt_exp = np.repeat(prtxprt_exp, rpt)

  return prtxprt_exp

def calcu_prtxprt_power(dpos_i, dpos_j, t_i, t_j):

  return np.power(dpos_i[:,0], quant[t_i,0]) * np.power(dpos_i[:,1], quant[t_i,1]) * np.power(dpos_i[:,2], quant[t_i,2]) * np.power(dpos_j[:,0], quant[t_j,0]) * np.power(dpos_j[:,1], quant[t_j,1]) * np.power(dpos_j[:,2], quant[t_j,2])

def calcu_coef(g_lgd, g_lgd_o, rpt):
  coef = np.divide(np.power(g_lgd,2), np.power(1-g_lgd_o,2))
  coef = np.repeat(coef, rpt)

  return coef

def calcu_prtxprt(grid, g_lgd, g_lgd_o, wgt, typ_i, typ_j, eps_ij, ks_ij, posc_i, posc_j, wposc_ij):
  dpos_i = np.subtract(grid, posc_i-wposc_ij)
  dpos_j = np.subtract(grid, posc_j-wposc_ij)
  
  # calculate number of spherical terms
  rpt = len(wgt)/len(g_lgd)

  # calculate exponent term
  prtxprt_exp = calcu_prtxprt_exp(g_lgd, rpt, eps_ij)

  # calculate polynomial terms
  prtxprt_power = calcu_prtxprt_power(dpos_i, dpos_j, typ_i, typ_j)

  # calculate integral coeficient  
  coef = calcu_coef(g_lgd, g_lgd_o, rpt)
  
  # calculate exponent term outside the integration
  out_exp = np.exp(-ks_ij*np.power(np.linalg.norm(posc_i-posc_j),2))

  # calculate prtxprt
  prtxprt = wgt * coef * prtxprt_exp * prtxprt_power * 4 * np.pi * out_exp
  
  return prtxprt

def quad_int_rg(prtxprt, grid, rg, wposc, num_prt):
  # transform rg coordinates
  rg = rg - wposc

  # to save the memory, we iterate it with row of num_prt
  ovlp_prt_rg = np.empty((num_prt, num_prt))
  for p in range(num_prt):
    # calculate r-rg, grid*3 num_prt*num_prt*grid*3
    r_rg_p = np.linalg.norm(grid-rg[p,:,np.newaxis,:], axis=-1)
    r_rg_p[r_rg_p == 0] = 1e-8

    # calculate overlap
    ovlp_prt_rg[p] = np.sum(np.divide(prtxprt[p], r_rg_p), axis=-1)

  return ovlp_prt_rg

def gpu_kernel(grid_gpu, sub_rg, sub_prtxprt_gpu, l):

  sub_rg_gpu = gpuarray.to_gpu(np.asarray(sub_rg, np.float32))

  # to save memory, we overwrite in each calculation
  # norm
  tmp_gpu = cumi.add_matvec(grid_gpu, -sub_rg_gpu)
  tmp_gpu = tmp_gpu.reshape(-1,3)
  tmp_gpu = cumi.multiply(tmp_gpu, tmp_gpu)
  tmp_gpu =  cumi.sum(tmp_gpu, axis=-1)
  tmp_gpu = pycuda_sqrt(tmp_gpu)

  tmp_gpu = culi.transpose(tmp_gpu.reshape(-1,l))

  tmp_gpu = cumi.sum(cumi.divide(sub_prtxprt_gpu, tmp_gpu), axis=-1)

  tmp = tmp_gpu.get()

  sub_rg_gpu.gpudata.free()
  tmp_gpu.gpudata.free()

  return tmp

def quad_int_rg_gpu(prtxprt_gpu, grid_gpu, rg, wposc, num_prt, ind, s):
  # transform rg coordinates
  # now rg is also num_prt*num_prt*3
  rg = rg - wposc
  rg = rg.reshape(-1,3)

  # calculate in gpu
  ovlp_prt_rg = gpu_kernel(grid_gpu, rg[ind[s]].reshape(-1), prtxprt_gpu, len(ind[s]))

  return ovlp_prt_rg

def split_array(prtxprt, spt, num_prt):
  ind = np.split(np.array(range(num_prt*num_prt)), spt)
  #ind3 = np.split(np.array(range(num_prt*num_prt*3)), spt)
  
  prtxprt_list = []
  for s in range(spt):
    tmp = prtxprt[ind[s]]
    prtxprt_list.append(tmp)

  return ind, prtxprt_list

def check_lumo(num_orb, noon):
  lumo = int(round(np.sum(noon)/2))
  if lumo == num_orb:
    print("Fully occupied, no need for MP2")
  
  return lumo

def calcu_eden_mp2(num_orb, lumo, noe, int2el, wfn, A):
  e = np.zeros((len(wfn)))
  for i in range(lumo):
    for j in range(lumo):
      for a in range(lumo,num_orb):
        for b in range(lumo,num_orb):
          ijab = int2el[i,j,a,b]
          ijba = int2el[i,j,b,a]
          e = e + (wfn[:,i]*wfn[:,a]*A[:,j,b]) * (2*ijab - ijba)/(noe[i]+noe[j]-noe[a]-noe[b])

  return e


fwfn = sys.argv[1]
ffci = sys.argv[2]
frg = sys.argv[3]
fsel = sys.argv[4]
fout = sys.argv[5]

# load wave functions from h5py
with h5py.File(fwfn, 'r') as hf:
  typ_orb, num_orb, num_prt, num_nuc = hf['BASIC'][:]
  pos = hf['POS'][:]
  cassign = hf['CASSIGN'][:]
  tassign = hf['TASSIGN'][:]
  exp = hf['EXP'][:]
  noon = hf['NOON'][:]
  noe = hf['NOE'][:]
  nocoef = hf['NOCOEF'][:]

# load 2-ele integration fcidump from npy
#int2el = np.load(ffci)

# load standard grid from npz
rg_grid = np.load(frg)['GRID']
rg_wfn = np.load(frg)['WFN']
rg_wgt = np.load(frg)['WGT']

# load index selection of rg grid
try:
  rg_sel = np.load(fsel)
except:
  try:
    num = int(fsel)
    if num == 0:
      rg_sel = range(len(rg_grid))
    else:
      rg_sel = range(len(num))
  except:
    start, end = map(int, fsel.split(','))
    rg_sel = range(start,end)

# fill the type with quantum number -> 35 * [n,m,l]
quant = quant_fill().astype(int)

# construct grid from Legendre quadrature Lebedev sphere quadrature
g_lgd_o, g_lgd, grid, wgt = gen_grid()

# rotate the grid to prevent overlap with rg

rx = rot3d.from_euler('x', 5, degrees=True)
ry = rot3d.from_euler('y', 5, degrees=True)
rz = rot3d.from_euler('z', 5, degrees=True)
r = rx * ry * rz
grid = r.apply(grid)

# combine two exponents into one based on the weighted position of center
posc = np.array([pos[cassign[i]-1] for i in range(num_prt)])
eps, ks, wposc = reduce_coord(posc, exp, num_prt)


typ = tassign - 1
# prepare for overlap for each point (prt_i(r)*prt_j(r)) for grid
prtxprt = np.empty((num_prt, num_prt, len(grid)), dtype=np.float32)
for i in range(num_prt):
  for j in range(i+1):
    prtxprt[i,j] = calcu_prtxprt(grid, g_lgd, g_lgd_o, wgt, typ[i], typ[j], eps[i,j], ks[i,j], posc[i], posc[j], wposc[i,j])
    prtxprt[j,i] = prtxprt[i,j]

# integrate primitive overlap at rg point (sum prt_i(r)*prt_j(r)/(r-rg))

# gpu
# wposc is num_prt*num_prt*3
# grid_ext is grid*(num_prt*num_prt*3)
# all in FP32
spt = 1

grid = np.asarray(grid, np.float32)
prtxprt = np.asarray(prtxprt, np.float32)
#grid_ext = np.tile(grid, (1,num_prt*num_prt)) # largest array 14600*3*np^2*4 nytes
grid_ext = np.tile(grid, (1, int(num_prt*num_prt/spt)))

#grid_ext = np.ascontiguousarray(grid_ext, dtype=np.float32)
prtxprt = prtxprt.reshape(-1, len(grid))

ind, prtxprt_list = split_array(prtxprt, spt, num_prt)

del grid
#del grid_ext
del prtxprt

culi.init()
ovlp_prt_rg_grid = np.empty((len(rg_sel), num_prt*num_prt), np.float32)
grid_gpu = gpuarray.to_gpu(grid_ext)
for s in range(spt):
  prtxprt_gpu = gpuarray.to_gpu(prtxprt_list[s])
  #grid_gpu = gpuarray.to_gpu(grid_list[s])
  ovlp_prt_rg_grid_s = []
  for gi in tqdm.tqdm(range(len(rg_sel))):
    g = rg_sel[gi]
    ovlp_prt_rg_grid[gi,ind[s]] = quad_int_rg_gpu(prtxprt_gpu, grid_gpu, rg_grid[g], wposc, num_prt, ind, s)
  prtxprt_gpu.gpudata.free()

  # release the memory by setting to a number
  prtxprt_list[s] = 0
  #grid_list[s] = 0

grid_gpu.gpudata.free()


ovlp_prt_rg_grid = ovlp_prt_rg_grid.reshape(-1, num_prt, num_prt)

'''
# mp
culi.init()
with multiprocessing.Pool(processes=1) as pool:
  ovlp_prt_rg_grid = pool.starmap(quad_int_rg_gpu, tqdm.tqdm([(prtxprt, grid, rg_grid[g], wposc, num_prt) for g in rg_sel]))
ovlp_prt_rg_grid = np.array(ovlp_prt_rg_grid)
'''

# transform to NO basis (c*ovlp*c^T)
ovlp_no_rg_grid = np.matmul(np.matmul(nocoef, ovlp_prt_rg_grid), nocoef.T)


# get the lumo, if fully occupied (lumo is num_orb) return error
lumo = check_lumo(num_orb, noon)

# load 2-ele integration fcidump from npy
int2el = np.load(ffci)
int2el = np.float32(int2el)

eden_mp2 = calcu_eden_mp2(num_orb, lumo, noe, int2el, rg_wfn[rg_sel], ovlp_no_rg_grid)

#print(eden_mp2)

np.save(fout, eden_mp2)
